import logging
import time

from PyQt5.QtCore import QObject, QJsonValue

from ... import mapwidget


class Evented(QObject):
    '''
    Base class for all pyqtlet2 objects.
    Handles initiation, as well as all python<->js communication
    '''
    mapWidget = None

    def __init__(self, mapWidget=None):
        '''
        Base class for all pyqtlet2 objects
        Handles initiation, as well as python-Js communication
        The first pyqtlet2 object to be initiated should be pyqtlet2.L.map
        This will allow all the pyqtlet2 objects to have access to the
        widget and thus the ability to implement leaflet via python.

        :param pyqtlet2.MapWidget mapWidget: The mapwidget object
            Should only be sent once, when the first object is being 
            initialised.
        '''
        super().__init__()
        self._logger = logging.getLogger(__name__)
        self._logger.setLevel(9)
        self.response = None
        if Evented.mapWidget:
            return
        if mapWidget is None:
            raise RuntimeError('L.map must be initialised before other pyqtlet2 objects')
        if not issubclass(type(mapWidget), mapwidget.MapWidget):
            raise TypeError(('Expected mapWidget of type pyqtlet2.MapWidget, '
                            'received {type_}'.format(type_=type(mapWidget))))
        Evented.mapWidget = mapWidget
        js = ('var channelObjects = null;'
              'new QWebChannel(qt.webChannelTransport, function(channel) {'
              '    channelObjects = channel.objects;'
              '});')
        self.runJavaScript(js)
        self.mapWidget.page.titleChanged.connect(lambda: print('title changed'))

    def getJsResponse(self, js, callback):
        '''
        Runs javascript code in the mapWidget and triggers callback.

        Can be used for custom use cases where information is required
        from the mapwidget, and the existing code does not cover the
        requirement
        
        :param str js: The javascript code
        :param function callback: The function that will consume the 
            javascript response

        .. note:: 
            Qt runs runJavaScript function asynchronously. So if we want 
            to get a response from leaflet, we need to force it to be sync
            In all that I have tried, I was unable to get the response from
            the same function, so I am converting it to a method with callback
        '''
        self._logger.debug('Running JS with callback: {js}=>{callback}'.format(
            js=js, callback=callback.__name__))
        self.mapWidget.page.runJavaScript(js, callback)

    def runJavaScript(self, js):
        '''
        Runs javascript code in the mapWidget.

        Can be used for custom use cases where the existing code,
        methods etc. do not cover the requirements.

        :param str js: The javascript code
        '''
        self._logger.debug('Running JS: {js}'.format(js=js))
        self.mapWidget.page.runJavaScript(js)

    def _createJsObject(self, leafletJsObject):
        '''
        Function to create variables/objects in leaflet in the
        javascript "engine", and registers the object so that it can
        be called in the webchannel.

        :param str leafletJsObject: javascript code that creates the
            leaflet object
        '''
        # Creates the js object on the mapWidget page
        js = 'var {name} = {jsObject}'.format(name=self.jsName, 
                jsObject=leafletJsObject)
        self.runJavaScript(js)
        # register the object in the channel
        self.mapWidget.channel.registerObject(
                '{name}Object'.format(name=self.jsName), self)

    def _connectEventToSignal(self, event, signalEmitter):
        # We need to delete some keys as they are causing circular structures
        js = '{name}.on("{event}", function(e) {{\
                  delete e.target;\
                  delete e.sourceTarget;\
                  e = copyWithoutCircularReferences([e], e);\
                  channelObjects.{name}Object.{signalEmitter}(e)}})'.format(
            name=self.jsName, event=event, signalEmitter=signalEmitter)
        self.runJavaScript(js)

    def _stringifyForJs(self, object_):
        # When passing options to JS, sometimes we need to pass in objects
        # this method and _handleObject take care of that
        # Some arguments are strings and some are objects. We also make sure
        # that the objects are not sent as strings. Similarly, we also convert
        # python bool to js bool etc.
        jsString = str(self._handleObject(object_))
        jsString = jsString.replace('\'__pyqtletObjectStart__', '')
        jsString = jsString.replace('\"__pyqtletObjectStart__', '')
        jsString = jsString.replace('__pyqtletObjectEnd__\'', '')
        jsString = jsString.replace('__pyqtletObjectEnd__\"', '')
        return jsString

    def _handleObject(self, object_):
        if type(object_) is list:
            return [self._handleObject(item) for item in object_]
        if type(object_) is dict:
            return {key: self._handleObject(object_[key]) for key in object_}
        if issubclass(object_.__class__, Evented):
            return '__pyqtletObjectStart__{name}__pyqtletObjectEnd__'.format(name=object_.jsName)
        if object_ is True:
            return '__pyqtletObjectStart__true__pyqtletObjectEnd__'
        if object_ is False:
            return '__pyqtletObjectStart__false__pyqtletObjectEnd__'
        if object_ is None:
            return '__pyqtletObjectStart__null__pyqtletObjectEnd__'
        return object_

    def _qJsonValueToDict(self, object_):
        # Qt returns QJsonValue from within the QChannel. Converting
        # that into a dict is a small recursive function.
        if type(object_) is QJsonValue:
            return self._qJsonValueToDict(self._qJsonToRespectiveType(object_))
        if type(object_) is list:
            return [self._qJsonValueToDict(item) for item in object_]
        if type(object_) is dict:
            return {key: self._qJsonValueToDict(object_[key]) for key in object_}
        return object_

    def _qJsonToRespectiveType(self, object_):
        # A QJsonValue can be one of many types. This function just
        # converts into the correct type
        if object_.isArray():
            return object_.toArray()
        if object_.isBool():
            return object_.toBool()
        if object_.isDouble():
            return object_.toDouble()
        if object_.isNull():
            return None
        if object_.isObject():
            return object_.toObject()
        if object_.isString():
            return object_.toString()
        if object_.isUndefined():
            return None




from ..layer import Layer

from PyQt5.QtCore import pyqtSlot, pyqtSignal, QJsonValue

class Marker(Layer):
    moveend = pyqtSignal(dict)
    move = pyqtSignal(dict)

    def __init__(self, latLng, options=None):
        super().__init__()
        self.latLng = latLng
        self.options = options
        self._initJs()
        self._connectEventToSignal('move', '_onMove')
        self._connectEventToSignal('moveend', '_onMoveend')

    @pyqtSlot(QJsonValue)
    def _onMove(self, event):
        self._logger.debug('map clicked. event: {event}'.format(event=event))
        event = self._qJsonValueToDict(event)
        self.latLng = [event["latlng"]["lat"], event["latlng"]["lng"]]
        self.move.emit(event)

    @pyqtSlot(QJsonValue)
    def _onMoveend(self, event):
        self._logger.debug('map clicked. event: {event}'.format(event=event))
        self.moveend.emit({**self._qJsonValueToDict(event), "sender": self})

    def _initJs(self):
        leafletJsObject = 'L.marker({latLng}'.format(latLng=self.latLng)
        if self.options:
            leafletJsObject += ', {options}'.format(options=self.options)
        leafletJsObject += ')'
        self._createJsObject(leafletJsObject)

    def setLatLng(self, latLng):
        js = '{layerName}.setLatLng({latLng})'.format(
                layerName=self._layerName, latLng=latLng)
        self.runJavaScript(js)

    def setOpacity(self, opacity):
        js = '{layerName}.setOpacity({latLng})'.format(
                layerName=self._layerName, opacity=opacity)
        self.runJavaScript(js)



from ..core import Evented

class Layer(Evented):

    # layerId is a static variable shared between all layers
    # It is used to give unique names to layers
    layerId = 0

    @property
    def layerName(self):
        return self._layerName

    @layerName.setter
    def layerName(self, name):
        self._layerName = name

    @property
    def jsName(self):
        return self._layerName

    @property
    def map(self):
        return self._map

    @map.setter
    def map(self, map_):
        self._map = map_

    def __init__(self):
        super().__init__()
        self._map = None
        self._layerName = self._getNewLayerName()

    def _getNewLayerName(self):
        layerName = 'l{}'.format(self.layerId)
        Layer.layerId += 1
        return layerName

    def addTo(self, map_):
        map_.addLayer(self)

    def removeFrom(self, map_):
        map_.removeLayer(self)

    def bindPopup(self, content, options=None):
        js = '{layerName}.bindPopup("{content}"'.format(
                layerName=self._layerName, content=content)
        if options:
            js += ', {options}'.format(self._stringifyForJs(options))
        js += ')'
        self.runJavaScript(js)
        
    def unbindPopup(self):
        js = '{layerName}.unbindPopup()'.format(layerName=self._layerName)
        self.runJavaScript(js)

    def bindTooltip(self, content, options=None):
        js = '{layerName}.bindTooltip("{content}"'.format(
                layerName=self._layerName, content=content)
        if options:
            js += ', {options}'.format(self._stringifyForJs(options))
        js += ')'
        self.runJavaScript(js)
        
    def unbindTooltip(self):
        js = '{layerName}.unbindTooltip()'.format(layerName=self._layerName)
        self.runJavaScript(js)





from setuptools import setup

OPTIONS={
    "py2app": {
        "resources": ["web"]
    }
}

setup(
    name='pyqtlet2',
#    app=["playground.py"],
    version='0.4.1',
    description='Bringing leaflet maps to PyQt',
    author='Leon Friedmann',
    author_email='leon.friedmann@tum.de',
    options=OPTIONS,
    url='https://github.com/JaWeilBaum/pyqtlet2',
    keywords=['leaflet', 'pyqt', 'maps', 'python', 'python3'],
    classifiers=[],
)



import os
import time

from PyQt5.QtCore import QEventLoop, QObject, QUrl, pyqtSignal
from PyQt5.QtWebChannel import QWebChannel
from PyQt5.QtWebEngineWidgets import ( QWebEngineView, QWebEnginePage, QWebEngineSettings, 
                                       QWebEngineScript )


class MapWidget(QWebEngineView):
    """
    The MapWidget class is a QWebEngineView that houses the leaflet map.
    Since it is a QWidget, it can be added to any QLayout.
    """

    @property
    def page(self):
        return self._page

    @property
    def channel(self):
        return self._channel

    def __init__(self):
        super().__init__()
        self._page = QWebEnginePage()
        self.setPage(self._page)
        self._channel = QWebChannel()
        self._page.setWebChannel(self._channel)
        self._loadPage()

    def _get_page_path(self):
        return os.path.join(os.getcwd(), 'web', 'map.html')

    def _loadPage(self):
        html_path = self._get_page_path()
        # QEventLoop is used to make the page loading behave syncronously
        init_loop = QEventLoop()
        self._page.loadFinished.connect(init_loop.quit)
        self._page.load(QUrl().fromLocalFile(html_path))
        init_loop.exec()

